/*  Tree construction for Greenspan/Barnes-Hut N-body calculation.	*/
/*  Source file:  tree.cs	Prototypes:  tree.hs			*/
/*  Uses header  gbh.hs  which defines shapes Tree and Particle,	*/
/*  external variables associated with those shapes, and aliases for	*/
/*  the de-referenced forms which are not arrays.			*/

#include <stdio.h>
#include <math.h>
#include <cscomm.h>
#include "gbh.hs"


/*  Interactive query for tree construction option	*/
/*  Returns option code.				*/

int Toption(void)
 { int tcode;
  static char m0[]=" box", m1[]=" cube (centered)";
  static char m2[]=" cube (off-center)";

   printf("Tree: 0 box, 1 centered cube, 2 offcenter cube, -1 none > ");
   scanf("%d", &tcode);
    if ( (0 <= tcode) & (tcode <= 2) )
      printf("Bounding region is a%s.\n",
	     (tcode==0) ? m0 : (tcode==1) ? m1 : m2);
     else tcode = -1;
   return (tcode);
 }


/*  Print summary description of the tree.	*/

void Tsummary(void)
 { with ( Tree ) everywhere
    { printf("Tree characteristics:\n");
      printf("  Boundary   x:  %13.5e  %13.5e\n", Ba[0], Bb[0]);
      printf("             y:  %13.5e  %13.5e\n", Ba[1], Bb[1]);
      printf("             z:  %13.5e  %13.5e\n", Ba[2], Bb[2]);
      printf("  Interior nodes:  %d    Leaves:  %d    Total nodes:  %d\n",
			      Nodes-N,		 N,		   Nodes  );
      printf("  Construction levels:  %d    Height:  %d\n", T, HT);
    }
 }


/*  Print complete tree.  Begins by printing summary.  If `data' then	*/
/*  Xp0, Xp1 must be supplied, otherwise CMC_no_field is OK for these.	*/

void Tdetail(int data, float:Tree Xp0[3], float:Tree Xp1[3],
		       float:Tree *pmp,   float:Tree *pTh    )
 { int u, p;
   Tsummary();
   printf("Node Parent Child0 Chldrn");
   if ( data )
      printf("   mass / h  -------------  Xp0 / Xp1  -------------\n");
    else printf(" Next      mass            h\n");
   for ( u = 0; u < Nodes; u++ )
    { p = [u]Parent;  if ( p == moo ) p = -9999;
      printf("%4d  %5d  %4d  %4d", u, p, [u]Child0, [u]Chldrn);
      if ( data)
       { printf("  %12.5e %12.5e %12.5e %12.5e\n",
		 [u]*pmp, [u]Xp0[0], [u]Xp0[1], [u]Xp0[2]);
	 printf("      Next:  %4d        %12.5e %12.5e %12.5e %12.5e\n",
		   [u]Tnext, [u]*pTh, [u]Xp1[0], [u]Xp1[1], [u]Xp1[2]);
       }
       else printf("  %4d  %12.5e  %12.5e\n", [u]Tnext, [u]*pmp, [u]*pTh);
    }	/* end for */
 }


/*  Print particle cells given by c.  Can be called anywhere.	*/

void Pcells(int:Particles *c)
 { int u, j, cols;
   cols = N <? 4;			/* Number of columns	*/
   printf("\nParticle  Cell");
   for ( u = 1; u < cols; u++ ) printf("     Particle  Cell");
   u = 0;				/* Particle number	*/
   while ( u < N )
    { printf("\n");
      for ( j = 0; ; )	/* Exits with break	*/
       { printf("%6d    %4d", u, [u]*c);  u++;  j++;
	 if ( (j == 4) | (j == N) ) break;
	 printf("     ");
       }
    }
   printf("\n");
 }


/*  Calculate tree "top down", returning indices of leaf nodes. */
/*  Must be called from Particles with all vp's active.		*/
/*  Xc specifies particle centers.				*/
/*    cube = 0:	use bounding box.				*/
/*    cube = 1: use centered cube.				*/
/*    cube = 2: use off-center cube retaining limits of		*/
/*		smaller magnitude.				*/

int:Particles TreeTop(float:Particles Xc[3], int cube)
 { int:Tree Pcount,			/* Number of particles in cell	*/
	      last,			/* Last child			*/
	        Ti;			/* Coordinate			*/
   bool:Tree treeseg;			/* For segmented scans		*/
   float:Particles Ca[3], Cb[3],	/* Lower, upper limits for cell	*/
		   md;			/* working			*/
   int:Particles subc, dsubc, actv;	/* Subcell, decoded, active	*/
   float ua, ub, um, ux;		/* For bounding box		*/
   int   ucells, u,			/* For level-1 cells, loops	*/
               mxu,			/* Coord with longest edge	*/
             Tlast,			/* Position of last tree entry	*/
	    ncells;			/* Number of cells in level	*/

 everywhere	/* in Particles */
 { um = 0.0;				/* For max edge	*/
   for ( u = 0;  u < 3;  u++ )		/* Calculate bounding box	*/
    { Ba[u] = ua = (<?= Xc[u]);  Bb[u] = ub = (>?= Xc[u]);	/* limits */
      ux = ub - ua;  if ( ux > um ) { um = ux;  mxu = u; }	/* max edge */
    }
   if ( cube == 1 )			/* Adjust to centered cube	*/
    for ( u = 0; u < 3; u++ )
     if ( u != mxu )
      { ux = 0.5*(um - (Bb[u] - Ba[u]));  Ba[u] -= ux;  Bb[u] += ux; }
   if ( cube == 2 )		/* Adjust to off-center cube	*/
    for ( u = 0; u < 3; u++ )
     if ( u != mxu )
      { ua = Ba[u];  ub = Bb[u];  ux = (um - (ub - ua));
	if ( fabs(ub) > fabs(ua) ) Bb[u] += ux; else Ba[u] -= ux;
      }
   /* Calculate level-1 subcells */
   subc = 0;				/* Will become subcell number	*/
   for ( u = 0;  u < 3;  u++ )
    { Ca[u] = ua = Ba[u];  Cb[u] = ub = Bb[u];
      um = 0.5f*(ua + ub);  subc <<= 1;		/* middle */
      where ( Xc[u] > um ) { subc++;  Ca[u] = um; } else Cb[u] = um;
    }
   dsubc  = 1 << subc;			/* Decoded subcell		*/
   ucells = (|= dsubc);			/* Occupied level-1 cells	*/
   /* printf(" ucells(hex):  %x\n", ucells);	    /*** Testing ***/
   cell   = 1 + subc;  ncells = 0;	/* To from cell addresses	*/
   for ( u = 0;  u < 7; u++ )		/* Calculate subcell address	*/
    { if ( ucells&1 ) ncells++; else where ( subc > u ) cell--;
      ucells >>= 1;			/* and count subcells		*/
    }
   if ( ucells ) ncells++;		/* Count last subcell		*/
   actv = 1;				/* All particles active		*/
   /* printf(" ncells:  %d\n", ncells);	    /*** Testing ***/

   /* Initialize tree, level-1 cells */
   with ( Tree ) everywhere
    { Parent = Child0 = Chldrn = Pcount = 0;
      Ti = pcoord(0);  Tnext = Ti + 1;		/* "Normal" Tnext	*/
      Tlast = ncells;  T = 1;
      [0]Parent = moo;  [0]Tnext = -1;  [0]Child0 = 1;  [0]Pcount = ncells;
      [Tlast]Tnext = -1;
    }
   T = 1;		/* One tree level constructed */

   /* Loop to finish tree.  For each active particle:			*/
   /*   cell = address of current cell, may or may not be leaf		*/
   /*   Ca, Cb delimit cell containing the particle			*/
   /* There are some active particles.					*/
   /* Pcount = Chldrn = Child0 = 0 in active and unused tree nodes	*/
   /*   except nodes which fail to split have Pcount = 1		*/
   /* Pcount = no. of children, Chldrn = set of subcells, and		*/
   /*   Child0 > 0 in nodes that have already split			*/
   /* Pcount = 1, Chldrn = Child0 = 0 in leaves				*/
   /* Particle shape is selected everywhere				*/

   for (;;)		/* Exits by break  */
    { where ( actv )			
       { [cell]Pcount += (int:current) 1;	/* Count particles	*/
	 subc = [cell]Pcount;
	 where ( subc == 1 ) actv = 0;
       }
    /*  Nodes = Tlast + 1;			    /*** Testing ***/
    /*  Tdetail(0, CMC_none, CMC_none, CMC_none);	    /*** Testing ***/
    /*  Pcells(&cell);				    /*** Testing ***/
      if ( (T > N) | !(|= actv) ) break;	/* Exit if none active	*/
      /* Calculate subcells */			/* T > N for errors	*/
      subc = 0;				/* Will become subcell number	*/
      for ( u = 0;  u < 3;  u++ )
       { md = 0.5f*(Ca[u] + Cb[u]);  subc <<= 1;	/* middle */
	 where ( Xc[u] > md ) { subc += 1;  Ca[u] = md; }
		     else     { Cb[u] = md; }
       }
      dsubc = 1 << subc;		/* Decoded subcell		*/
      where ( actv )			/* Find occupied subcells	*/
        [cell]Chldrn |= dsubc;

      /* Expand tree to next level */
      T++;				/* Count level	*/
      with ( Tree ) everywhere
      { where ( (Child0 == 0) & (Chldrn > 0) )	/* Active nodes		*/
        { Pcount = 0;  Child0 = Chldrn;   	/* To count subcells	*/
	  for ( u = 0; u < 8; u++ )
	   { where ( Child0&1 ) Pcount++;  Child0 >>= 1; }
	  /* Now leaves have Pcount = 1, Child0 = 0 */
	  where ( Pcount > 1 ) 				/* Cell splits */
	   { Chldrn |= 256;  ncells = (+= Pcount);
	     Child0 =  Tlast
		     + scan(Pcount, 0, CMC_combiner_add, CMC_upward,
			    CMC_none, CMC_no_field, CMC_exclusive  );
	     last = Child0 + Pcount;
	     [last]Parent = Ti;		/* Last child		*/
	     [last]Tnext = Tnext;	/* Next for last child	*/
	     Child0++;					/* 0-th child	*/
	   }
	  else Chldrn = 0;				/* For next cycle */
	}  /* end where (active nodes)	*/
       treeseg = Parent > 0;				/* Fill in parents */
       where ( Ti > Tlast )				/* New nodes	   */
	 Parent = scan(Parent, 0 , CMC_combiner_copy, CMC_downward,
			    CMC_start_bit, &treeseg, CMC_inclusive );
       Tlast += ncells;
      }  /* end with ( Tree ) */
      where ( actv )
       { dsubc = [cell]Chldrn;				/* New subcells	*/
	 where ( dsubc & 256 )				/* Cell divided	*/
	  { cell = [cell]Child0 + subc;			/* New cell	*/
	    for ( u = 0; u < 7; u++ )	/* Adjust cell address	*/
	     { where ( (subc > u) & !(dsubc&1) ) cell--;
	       dsubc >>= 1;
             }
          }
       }
    }	/* End tree-construction loop	*/

   /* Finish tree representation */
   with ( Tree ) everywhere
    { where ( Pcount  > 1 ) Chldrn = Pcount;
      where ( Pcount == 1 ) Chldrn = 0;
    }
 }	/* end everywhere in Particles */
   Nodes = Tlast+1;			/* Final number of nodes */
   return ( cell );
 }	/* end TreeTop */
